from typing import Optional
from pydantic import BaseModel, Field, SecretStr, field_validator, ConfigDict


class Payload(BaseModel):
    sub: str
    email_verified: Optional[bool]
    iss: str
    username: str
    origin_jti: str
    aud: str
    event_id: str
    token_use: str
    auth_time: str
    exp: str
    iat: str
    jti: str
    email: Optional[str]
    preferred_username: Optional[str] = None

    model_config = ConfigDict(coerce_numbers_to_str=True)


class AuthenticationResult(BaseModel):
    access_token: str = Field(title="アクセストークン")
    id_token: str = Field(title="IDトークン")
    refresh_token: str = Field(title="更新トークン")
    token_type: str = Field(title="トークンタイプ")
    expires_in: int = Field(title="トークンが有効な秒数")


class RefreshTokenResult(BaseModel):
    access_token: str = Field(title="アクセストークン")
    id_token: str = Field(title="IDトークン")
    token_type: str = Field(title="トークンタイプ")
    expires_in: int = Field(title="トークンが有効な秒数")


class SignupRequest(BaseModel):
    username: str = Field(title="ユーザー名")
    password: SecretStr = Field(title="パスワード", min_length=4)


class LoginRequest(BaseModel):
    username: str = Field(title="ユーザー名")
    password: SecretStr = Field(title="パスワード", min_length=4)


class ConfirmSignupRequest(BaseModel):
    username: str = Field(title="ユーザー名")
    confirmation_code: str = Field(title="検証コード", min_length=6, max_length=6)

    @field_validator("confirmation_code")
    def confirmation_code_validator(cls, v: str) -> str:
        if not v.isnumeric():
            raise ValueError("検証コードは6桁の数字です")
        return v


class ResendConfirmationCodeRequest(BaseModel):
    username: str = Field(title="ユーザー名")


class TokenRefreshRequest(BaseModel):
    refresh_token: str = Field(title="更新トークン")



from typing import List

import jwt
import requests
from pydantic import BaseModel


class Key(BaseModel):
    alg: str
    e: str
    kid: str
    kty: str
    n: str
    use: str


class Config:
    region: str
    pool_id: str
    client_id: str
    cognito_url: str
    cognito_jwk_url: str
    public_keys: List[Key] = []

    def __init__(
        self,
        *,
        region: str,
        pool_id: str,
        client_id: str,
    ) -> None:
        self.region = region
        self.pool_id = pool_id
        self.client_id = client_id

        url_format = "https://cognito-idp.{}.amazonaws.com/{}"
        self.cognito_url = url_format.format(
            self.region,
            self.pool_id,
        )
        self.cognito_jwk_url = f"{self.cognito_url}/.well-known/jwks.json"

        self._init_public_keys()

    def get_public_key(self, kid: str) -> str:
        for key in self.public_keys:
            if key.kid == kid:
                return jwt.algorithms.RSAAlgorithm.from_jwk(key.model_dump_json())
        else:
            raise Exception("not found public key")

    def _init_public_keys(self) -> None:
        res = requests.get(self.cognito_jwk_url)
        for key in res.json()["keys"]:
            self.public_keys.append(Key(**key))



from fastapi import HTTPException, status


class UserNotConfirmedException(Exception):
    pass


class NotAuthorizedException(Exception):
    pass


class UsernameExistsException(Exception):
    pass


NotAuthorizedHTTPException = HTTPException(
    status_code=status.HTTP_401_UNAUTHORIZED,
    detail="un authorized",
    headers={"WWW-Authenticate": "Bearer"},
)



from fastapi import Security
from fastapi.security import APIKeyHeader
from fastapi.security.utils import get_authorization_scheme_param

api_key = APIKeyHeader(name="Authorization")


def get_api_token(authorization: str = Security(api_key)) -> str:
    _, token = get_authorization_scheme_param(authorization)
    return token



from typing import Any, Dict

import boto3
import jwt
from fastapi import Depends

from cognito_pyauth import exceptions, schemas
from cognito_pyauth.config import Config
from cognito_pyauth.utils import get_api_token
from cognito_pyauth.logger import logger


class Auth:
    config: Config
    client: Any

    def __init__(self, config: Config) -> None:
        self.config = config
        self.client = boto3.client(
            "cognito-idp",
            region_name=self.config.region,
        )

    def get_payload(
        self,
        token: str = Depends(get_api_token),
    ) -> schemas.Payload:
        try:
            jwt_header = jwt.get_unverified_header(token)
            jwt_algorithms = jwt_header["alg"]
            public_key = self.config.get_public_key(jwt_header["kid"])
            decoded = jwt.decode(
                token,
                public_key,
                algorithms=[jwt_algorithms],
                verify=True,
                options={"require_exp": True},
                audience=self.config.client_id,
            )
            return schemas.Payload(
                **decoded,
                username=decoded["cognito:username"],
            )
        except Exception as e:
            logger.info(e)
            raise exceptions.NotAuthorizedException()

    def get_payload_depends(
        self,
        token: str = Depends(get_api_token),
    ) -> schemas.Payload:
        try:
            return self.get_payload(token)
        except exceptions.NotAuthorizedException:
            raise exceptions.NotAuthorizedHTTPException

    def login(
        self,
        username: str,
        password: str,
    ) -> schemas.AuthenticationResult:
        req = schemas.LoginRequest(
            username=username,
            password=password,
        )

        try:
            res = self.client.initiate_auth(
                AuthFlow="USER_PASSWORD_AUTH",
                AuthParameters={
                    "USERNAME": req.username,
                    "PASSWORD": req.password.get_secret_value(),
                },
                ClientId=self.config.client_id,
            )
        except self.client.exceptions.UserNotConfirmedException as e:
            raise exceptions.UserNotConfirmedException(e)
        except self.client.exceptions.NotAuthorizedException as e:
            raise exceptions.NotAuthorizedException(e)

        result = res["AuthenticationResult"]
        return schemas.AuthenticationResult(
            access_token=result["AccessToken"],
            id_token=result["IdToken"],
            refresh_token=result["RefreshToken"],
            token_type=result["TokenType"],
            expires_in=result["ExpiresIn"],
        )

    def signup(
        self,
        username: str,
        password: str,
    ) -> Dict[str, Any]:
        req = schemas.SignupRequest(
            username=username,
            password=password,
        )

        try:
            return self.client.sign_up(
                Username=req.username,
                Password=req.password.get_secret_value(),
                ClientId=self.config.client_id,
            )
        except self.client.exceptions.UsernameExistsException as e:
            raise exceptions.UsernameExistsException(e)

    def confirm_signup(
        self,
        username: str,
        confirmation_code: str,
    ) -> Dict[str, Any]:
        req = schemas.ConfirmSignupRequest(
            username=username,
            confirmation_code=confirmation_code,
        )
        return self.client.confirm_sign_up(
            Username=req.username,
            ConfirmationCode=req.confirmation_code,
            ClientId=self.config.client_id,
        )

    def resend_confirmation_code(
        self,
        username: str,
    ) -> Dict[str, Any]:
        req = schemas.ResendConfirmationCodeRequest(
            username=username,
        )
        return self.client.resend_confirmation_code(
            Username=req.username,
            ClientId=self.config.client_id,
        )

    def delete_user(
        self,
        username: str,
        password: str,
    ) -> Dict[str, Any]:
        res = self.login(username, password)
        return self.client.delete_user(
            AccessToken=res.access_token,
        )

    def delete_user_by_access_token(
        self,
        access_token: str,
    ) -> Dict[str, Any]:
        return self.client.delete_user(
            AccessToken=access_token,
        )

    def refresh_token(
        self,
        refresh_token: str,
    ) -> schemas.RefreshTokenResult:
        req = schemas.TokenRefreshRequest(refresh_token=refresh_token)

        res = self.client.initiate_auth(
            AuthFlow="REFRESH_TOKEN_AUTH",
            AuthParameters={
                "REFRESH_TOKEN": req.refresh_token,
            },
            ClientId=self.config.client_id,
        )
        result = res["AuthenticationResult"]
        return schemas.RefreshTokenResult(
            access_token=result["AccessToken"],
            id_token=result["IdToken"],
            token_type=result["TokenType"],
            expires_in=result["ExpiresIn"],
        )
